/*
 *      Copyright Richard Gorton, 2007-2011 All rights reserved
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 */

/*
 * compserver.c - A DWPPP (Dynamic Whole Program Path Profiler) compression server.
 *      takes advantage of processor dies having at least 2 cores.
 *      The incoming data is comprised of strata numbers.
 *      The server will compress the incoming content and write out the result
 *      to a pair of data files: a text file, and a binary file.
 *
 *      Written to compensate for suck-oh performance when all processing is
 *      done inside a PIN tool (and because multi-threaded PIN tools don't seem
 *      to work particularly well) as of the original (2007) construction date.
 *
 * Assumptions:
 *      Only one sender of data at a time
 *      Uses sysV shmem (file backed) as the data transfer method
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <zlib.h>

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include "servercomm.h"
/*
 * Organization of the resultant data:
 *      A text file containing entries:                 <dataFileName>.index
 *              insn epoch:path epoch:strata epoch:offset in the file
 *      A binary file containing the compressed data    <dataFileName>.data
 */
/*
 * server options: (not fully supported)
 *      -nowrite        do not write to the output file I/O (do compression)
 */

static struct DF_INFO   dataPair;
static struct SHARED_MEM*       s_sharedMem = NULL;
struct SHARED_MEM       g_localSHM;

static void cleanup()
{
}

static void cleanup_exit()
{
    fprintf (stderr, "Aborting\n");
    cleanup();
    exit (1);
}

static void processArgs (int argc, char *argv[])
{
}

/*
 * Simply compress the data, and write the records
 *      index file:     insn epoch:path epoch:strata epoch:data file offset
 *      data file:      just compressed raw data.  Format of the raw data determined by
 *                      strata tracker.
 */
void processPayload (struct SHARED_MEM* memData)
{
    struct PATH_MSG_HDR*  pHdr;
    uint64_t    cBufSize;
    char        idxBuf[1024];
    int         iBLen;
    char*       rawData;
    char        compressBuf[DB_BUF_SIZE];
    int         res;

    pHdr = (struct PATH_MSG_HDR*) memData;
        // Write out the index 
    iBLen = snprintf (idxBuf, sizeof(idxBuf) - 1, "%ld:%ld:%ld:%ld\n",
                        pHdr->startInsnEpoch,
                        pHdr->startPathEpoch,
                        pHdr->startStrataEpoch,
                        dataPair.dataFileOffset);
    res = write (dataPair.index_fd, idxBuf, iBLen);

        // Compress the raw data
    rawData = (void*) &memData->rawDBBuf;
    cBufSize = sizeof(compressBuf);
    res = compress2 (compressBuf, &cBufSize, rawData, pHdr->payloadSize, dataPair.compression);
    if (res != 0) {
        fprintf (stderr, "compress2 of %d bytes @ level %d returns %d\n",
                pHdr->payloadSize, dataPair.compression, res);
    }

        // And write out compressed data
    res = write (dataPair.data_fd, compressBuf, cBufSize);
    dataPair.dataFileOffset += cBufSize;
}

int main(int argc, char *argv[])
{
    ssize_t     rdata;
    int         clientlen;
    int         rc;
    int         exitcleanly = 0;
    char        idxFName [PATH_MAX];
    char        dataFName [PATH_MAX];

    processArgs (argc, argv);
    signal (SIGINT, cleanup_exit);

    unlink (COMPRESSOR_MMAP_FILE);
    s_sharedMem = OpenMMapFile (argv[0], SERVER_OPEN_MMAPFILE, COMPRESSOR_MMAP_FILE);
    if (s_sharedMem == NULL) {
        cleanup_exit();
    }

    dataPair.state = STATE_NONE;
        // And mark region as ready for use
    MarkReadyToFillData(s_sharedMem);

    while (!exitcleanly) {

        WaitUntilDataFilled(s_sharedMem);

        // Data in the shared memory buffer, ready to process

        // Only relevant once deployed in the field (if at all) and the format changes
        if (s_sharedMem->dataHdr.version != THIS_VERSION) {
            fprintf (stderr, "%s: received data from an unsupported version client: %d\n",
                        argv[0], s_sharedMem->dataHdr.version);
            continue; 
        }

        /*
         * At this point, start trying to be as concurrent as possible.  Make a local
         * copy of the content, and operate on that, and then mark the shared memory
         * region as available to write new data into.  Assume that it will be almost
         * completely full, so simply copy the entire thing.  The only time it will be
         * sparse is on the final set of data.
         * Once the copy is made, be sure to _not_ reference the shared memory region
         */
        memcpy ((void*) &g_localSHM, (void*) s_sharedMem, sizeof (g_localSHM));
        MarkReadyToFillData(s_sharedMem);

        // Open operations
        if (g_localSHM.dataHdr.operation & (OP_OPEN_APPEND|OP_OPEN_TRUNC)) {
            int openFlags;
            int prot_644;
            int rc_index;
            int rc_data;
            int res;

            if (dataPair.state == STATE_OPENED) {
                if (strcmp (g_localSHM.dataHdr.dataFileBaseName, dataPair.baseFileName) == 0) {
                    fprintf (stderr, "%s: [%s] is already open... ignoring open op\n",
                                argv[0], g_localSHM.dataHdr.dataFileBaseName);
                    goto skip_open;
                } else {
                    fprintf (stderr,
                                "%s: Unable to open (append) to [%s] - another file pair is in use: [%s]\n",
                                argv[0], g_localSHM.dataHdr.dataFileBaseName, dataPair.baseFileName);
                }
                continue;
            }
            prot_644 = S_IRUSR|S_IWUSR | S_IRGRP | S_IROTH;
            openFlags = O_RDWR;

            if (g_localSHM.dataHdr.operation & OP_OPEN_APPEND) {
                openFlags |= O_CREAT|O_APPEND;
            }
            if (g_localSHM.dataHdr.operation & OP_OPEN_TRUNC) {
                openFlags |= O_CREAT|O_TRUNC;
            }

            snprintf (idxFName, sizeof(idxFName) - 1, "%s.idx", g_localSHM.dataHdr.dataFileBaseName);
            rc_index = open (idxFName, openFlags, prot_644);
            if (rc_index == -1) {
                fprintf (stderr, "%s: Error opening [%s]: errno: %d\n", argv[0], idxFName, errno);
                fprintf (stderr, "aborting\n");
                exit (1);
            } else {
                char    iHdr[256];
                int     hLen;

                hLen = snprintf (iHdr, sizeof(iHdr) - 1,
                                "Finish    :Finish    :Finish      :Starting\n"
                                "insn epoch:path epoch:strata epoch:data file offset\n");
                res = write (rc_index, iHdr, hLen);
            }

            snprintf (dataFName, sizeof(dataFName) - 1, "%s.data", g_localSHM.dataHdr.dataFileBaseName);
            rc_data = open (dataFName, openFlags, prot_644);
            if (rc_data == -1) {
                fprintf (stderr, "%s: Error opening [%s]: errno: %d\n", argv[0], dataFName, errno);
                fprintf (stderr, "aborting\n");
                exit (1);
            }

            snprintf (dataPair.baseFileName, sizeof (dataPair.baseFileName) - 1, "%s", g_localSHM.dataHdr.dataFileBaseName);
            snprintf (dataPair.indexFileName, sizeof (dataPair.indexFileName) - 1, "%s.index", g_localSHM.dataHdr.dataFileBaseName);
            snprintf (dataPair.dataFileName, sizeof (dataPair.dataFileName) - 1, "%s.data", g_localSHM.dataHdr.dataFileBaseName);
            dataPair.index_fd = rc_index;
            dataPair.data_fd = rc_data;
            dataPair.dataFileOffset = 0x0;
            dataPair.state = STATE_OPENED;
            dataPair.flags = g_localSHM.dataHdr.flags;

                // Set libz compression level - default to 6
            if (g_localSHM.dataHdr.flags & FLAG_NOCOMPRESS) {
                dataPair.compression = 0;
            } else if (g_localSHM.dataHdr.flags & FLAG_COMPRESS_9) {
                dataPair.compression = 9;
            } else {
                dataPair.compression = 6;
            }
        }

skip_open: ;
        // Process the data payload (if any)
        processPayload (&g_localSHM);

        // Cleanup (if requested)
        if (g_localSHM.dataHdr.operation & (OP_FLUSH_CLOSE | OP_DELETE)) {
            close (dataPair.index_fd);
            close (dataPair.data_fd);
            dataPair.dataFileOffset = 0x0;
            dataPair.state = STATE_CLOSED;
        }

        if (g_localSHM.dataHdr.operation & OP_DELETE) {
            unlink (dataPair.indexFileName);
            unlink (dataPair.dataFileName);
            dataPair.indexFileName[0] = 0x0;
            dataPair.dataFileName[0] = 0x0;
            dataPair.dataFileOffset = 0x0;
            dataPair.state = STATE_NONE;
        }
    }

done: ;

    cleanup();
    exit (0);
}
