/*
 *      Copyright Richard Gorton, 2007-2011 All rights reserved
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 */
/*
 * Utility to extract paths and strata from the stored db's and dump them
 * based upon the path number.
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <zlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#include <string>
#include <vector>

extern "C" {
#include "puredb/puredb_read.h"
}
#include "pathprof.h"

#define ANALYZE_PATHS_MODE      0x1
#define ANALYZE_STRATA_MODE     0x2
#define VERBOSE_MODE            0x4
int     g_mode = 0;

typedef std::vector<int>    pathID_T;
pathID_T    g_pathIDvector;

#define PATHLOG_EXT     ".pathlog.txt"
#define STRATALOG_EXT   ".stratalog.txt"
#define PATHDB_EXT      ".pathinfo_db.p_dbp"
#define STRATADB_EXT    ".stratainfo_db.s_dbp"
#define STRATACONTENT_DATA      ".rawstrata.data"
#define STRATACONTENT_INDEX     ".rawstrata.idx"

std::string g_baseFile;
std::string *g_pathLog = NULL;
std::string *g_strataLog = NULL;
std::string *g_pathDB = NULL;
std::string *g_strataDB = NULL;
std::string *g_rawstrata_data = NULL;
std::string *g_rawstrata_index = NULL;

// palindromic isomers DO actually occur!
// Quite the surprise, indeed
// Move this declaration into a common header...
typedef struct ISOMER_STATS {
    int         numberPaths;    // number of unique paths at this length
    int         numIsomers;     // unique isomers of said paths
    int         numRotationals; // how many are rotational isomes
    int         numPalindromes; // Possible, but seems highly unlikely
            // dynamic counts at this path/strata length
    unsigned long   dynTotal;   // all paths
    unsigned long   dynIsomers;
    unsigned long   dynRotational;
    unsigned long   dynPalindrome;
    } ISOMER_STATS;
struct PSTATS   g_pathStats;
struct ISOMER_STATS     g_uniquePaths[VLPATHLIM];
void        *g_rawPaths[VLPATHLIM];
void        *g_canonPaths[VLPATHLIM];
int         g_sortVecLen;       // Needed to sort

typedef struct  PATHINFO {
    unsigned long   tripCount;
    bool            isIsomer;
    bool            isRotational;
    bool            isPalindrome;
    } PATHINFO;
struct PATHINFO *g_pathInfo = NULL;

static PureDB   pathDB_pureR;

void
Usage(char *argv0)
{
    fprintf (stderr, "Usage: %s <file basename> <pathid_list>\n", argv0);
    fprintf (stderr, "\tA tool to dump specific path ids collected form a whole program profile run\n");
    fprintf (stderr, "\twhere <file basename> is <executable name>.<pid> of the output from the wppp_3tier tool\n");
    fprintf (stderr, "\t<pathid_list> is a series of pathid numbers, separated by whitespace\n");
    fprintf (stderr, "\tFor example: if the output files are whet.5686.*, then the file basename is whet.5686\n");
}

static uint64_t
ReadLongFromPathDB(char *keystr)
{
    const char  *func = "ReadLongFromPathDB";
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    uint64_t    fetchVal;
    void        *dbReadp;
    int rc;
    uint64_t    retval = 0;
    char        uncompBuf[1024 * 1024];
    uint64_t    uncompLen = sizeof (uncompBuf);

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    Key.dptr = dbkey;
    Key.dsize = strlen(dbkey);
    Value.dptr = (char *)&fetchVal;
    Value.dsize = sizeof(fetchVal);
    
        // pathDB assumed to be open
    rc = puredb_find (&pathDB_pureR,
                       Key.dptr, Key.dsize,
                       (off_t *)&Value.dptr, (size_t *)&Value.dsize);
    if (rc != 0) {
        fprintf (stderr, "Problem in [%s] for key [%s]\n", func, keystr);
    } else {
        dbReadp = puredb_read (&pathDB_pureR, (off_t) Value.dptr, Value.dsize);
        retval = *(uint64_t *)dbReadp;
        puredb_read_free (dbReadp);
    }

    return retval;
}

static bool
ReadStructFromPathDB(char *keystr, void *valBuf, uint64_t valBufSize)
{
    const char  *func = "ReadStructFromPathDB";
    bool        success = true;
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    void        *dbReadp;
    int         rc;

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    Key.dptr = dbkey;
    Key.dsize = strlen(dbkey);
    Value.dptr = (char *)valBuf;
    Value.dsize = valBufSize;
    
        // pathDB assumed to be open
    rc = puredb_find (&pathDB_pureR,
                       Key.dptr, Key.dsize,
                       (off_t *)&Value.dptr, (size_t *)&Value.dsize);
    if (rc != 0) {
        success = false;
        fprintf (stderr, "Problem in [%s] for key [%s]\n", func, keystr);
    } else {
        dbReadp = puredb_read (&pathDB_pureR, (off_t) Value.dptr, Value.dsize);
        memcpy (valBuf, dbReadp, valBufSize);
        // retval = *(uint64_t *)dbReadp;
        puredb_read_free (dbReadp);
    }

    return success;
}

void
dumpAPath(int thisPathID)
{
    const char  *func = __FUNCTION__;
    int         rc;
    uint64_t    keyver;
    uint64_t    numpaths;
    bool        readOK;
    uint64_t    uniquePathLenMean;
    int         curPathSize;
    int         fillIndex[VLPATHLIM];

    rc = puredb_open (&pathDB_pureR, g_pathDB->c_str());
    if (rc != 0) {
        fprintf(stderr, "Weird failure opening [%s] - aborting\n", g_pathDB->c_str());
        exit(1);
    }

        // From paths.c:WritePathsDB()
        //      key             value
        //      KEY_VERNO       key version
        //      KEY_NUMPATHS    number paths
        //      KEY_PATHSTATS   path stat summary
        //      KEY_PATHID      path content

    keyver = ReadLongFromPathDB((char *)KEY_VERNO);
    if (keyver != KEY_VERSION) {
        fprintf(stderr, "Mismatch in db for key: [%s] got %ld\n", KEY_VERNO, keyver);
    }
    // fprintf(stdout, "key: [%s] got %d\n", KEY_VERNO, keyver);
    numpaths = ReadLongFromPathDB((char *)KEY_NUMPATHS);
    if (numpaths <= 100) {
        fprintf(stderr, "Probable error - only %ld paths - seems low\n", numpaths);
    }


    char                    dbkey[2048];
    struct MAX_PATH_ELT     maxPathBlob;
    struct PATH_QUEUE_ELT   corePathInfo;
    int                     pLen;
    datum       Key;
    datum       Value;

    memset (&maxPathBlob, 0, sizeof(maxPathBlob));
    snprintf (dbkey, sizeof(dbkey), "%s:%d", KEY_PATHID, thisPathID);
        // One BB_BLOB encapsulated in PATH_QUEUE_ELT
        // Read it twice (ugh) once to get the path length, and then a "full" read
        // This to ensure we do not read off from unmapped content
    ReadStructFromPathDB(dbkey, &corePathInfo, sizeof(corePathInfo));
    curPathSize =  (corePathInfo.pathLen - 1) * (sizeof (struct BB_BLOB))
                        + sizeof(corePathInfo);

    ReadStructFromPathDB(dbkey, &maxPathBlob, curPathSize);
    pLen = maxPathBlob.pathLen;

    // Header output, then each address & disassembly
    fprintf(stdout, "========== Path ID %d Disassembly ==========\n", thisPathID);
    for (int i = 0; i < pLen; i++) {
        char        KeyStr[128];
        int         KeyLen;
        char        *BbDisAsm;
        char        *CompressedDisAsm;

    // Buffer was only 1 Meg @ construction time
#define DISASM_BUF_SIZE (1024*1024)
        BbDisAsm = (char *) malloc(DISASM_BUF_SIZE);
        CompressedDisAsm = (char *) malloc(DISASM_BUF_SIZE);

        fprintf (stdout, "\t==== Bb Address: %p ====\n", maxPathBlob.dataBlob[i].BbAddress);
        snprintf (KeyStr, sizeof(KeyStr), "%p", maxPathBlob.dataBlob[i].BbAddress);
        KeyLen = strlen (KeyStr);
        Key.dptr = KeyStr;
        Key.dsize = KeyLen;
        Value.dptr = (char *)CompressedDisAsm;
        Value.dsize = DISASM_BUF_SIZE;
        rc = puredb_find (&pathDB_pureR,
                       Key.dptr, Key.dsize,
                       (off_t *)&Value.dptr, (size_t *)&Value.dsize);
        if (rc != 0) {
            fprintf (stderr, "Problem in [%s] for key [%s]\n", __FUNCTION__, KeyStr);
        } else if (Value.dsize >= DISASM_BUF_SIZE) {
            fprintf (stderr, "Problem in [%s] for Bb [%s] - too long!\n", __FUNCTION__, KeyStr);
        } else {
            CompressedDisAsm = (char *)puredb_read (&pathDB_pureR, (off_t) Value.dptr, Value.dsize);
            // And uncompress it
            uLongf  uncompSize = DISASM_BUF_SIZE;
            rc = uncompress((Bytef*)BbDisAsm, &uncompSize, (const Bytef*) CompressedDisAsm, Value.dsize);
            
            puredb_read_free (CompressedDisAsm);
        }
        fprintf (stdout, "%s\n", BbDisAsm);

        free (BbDisAsm);
    }
    fprintf(stdout, "========== End of Path ID %d Disassembly ==========\n", thisPathID);

    (void) puredb_close(&pathDB_pureR);

}

bool
fileExists(const char *pathName)
{
    struct stat stbuf;
    int         rc;
    bool        retval = false;

    rc = stat(pathName, &stbuf);
    if (rc == 0) {
        retval = true;
    }

    return retval;
}

bool
dataFilesOK(std::string &fname)
{
    bool        result = true;

    g_pathLog           = new std::string(fname + PATHLOG_EXT);
    g_strataLog         = new std::string(fname + STRATALOG_EXT);
    g_pathDB            = new std::string(fname + PATHDB_EXT);
    g_strataDB          = new std::string(fname + STRATADB_EXT);
    g_rawstrata_data    = new std::string(fname + STRATACONTENT_DATA);
    g_rawstrata_index   = new std::string(fname + STRATACONTENT_INDEX);

    if (!fileExists(g_pathLog->c_str())) {
        fprintf (stderr, "%s missing\n", g_pathLog->c_str());
        result = false;
    }
    if (!fileExists(g_strataLog->c_str())) {
        fprintf (stderr, "%s missing\n", g_strataLog->c_str());
        result = false;
    }
    if (!fileExists(g_pathDB->c_str())) {
        fprintf (stderr, "%s missing\n", g_pathDB->c_str());
        result = false;
    }
    if (!fileExists(g_strataDB->c_str())) {
        fprintf (stderr, "%s missing\n", g_strataDB->c_str());
        result = false;
    }
    if (!fileExists(g_rawstrata_data->c_str())) {
        fprintf (stderr, "%s missing\n", g_rawstrata_data->c_str());
        result = false;
    }
    if (!fileExists(g_rawstrata_index->c_str())) {
        fprintf (stderr, "%s missing\n", g_rawstrata_index->c_str());
        result = false;
    }

done:
    return result;
}

// Usage: pathdump <db file base name> <pathid_list>
// at least one specific pathid is required.  Dumping all paths
// is not supported at present, as it is not useful.
// pathid_list is a series of path numbers, delimited by spaces
int
main(int argc, char* argv[])
{
    int rc;

    if (argc < 3) {
        Usage(argv[0]);
        exit(1);
    }

    g_mode = ANALYZE_PATHS_MODE;

    // Lazy, but this is a research tool
    // argv[1] is the basefile
    g_baseFile = argv[1];
        // A complete set of files for a wpp (whole program profile) is:
        //      whet.5686.pathlog.txt
        //      whet.5686.stratalog.txt
        //      whet.5686.pathinfo_db.p_dbp
        //      whet.5686.rawstrata.data
        //      whet.5686.rawstrata.idx
        //      whet.5686.stratainfo_db.s_dbp
        //
    if (!dataFilesOK(g_baseFile)) {
        exit(1);
    }

    for (int i = 2; i < argc; i++)
    {
        int     pNum;

        pNum = atoi(argv[i]);
        dumpAPath(pNum);
    }

    return 0;
}

