/* Copyright Richard Gorton, 2007-2011 All rights reserved */

/*
 * servercomm.h - shared definitions across all of the system components:
 *      The main DWPPP (dynamic whole program path profiler) collector tool,
 *      the strata analyzer, and the final compression bit of code.
 *
 * When all the compression is done in the PIN tool, the overhead ends up being
 * large enough to make a WPPP virtually unusable.  By writing this as a client-server
 * application, it is possible to take advantage of a second processor core.
 * With some further work, it might be possible (and prove advantageous) to further
 * parallelize the work to take advantage of even more processor cores.
 */

#ifndef _SERVERCOMM_H_
#define _SERVERCOMM_H_

/*
 * 1 meg payload chunks - use DB_BUF_SIZE at all shmem transition points.
 * That way, it is possible to pass epoch data all the way through.
 */
#define DB_BUF_SIZE     (4*1024*1024)
#define THIS_VERSION    1

#define STRATA_INFO_DB_NAME     ".stratainfo_db"
#define RAW_STRATA_NAME         ".rawstrata"
#define SERVER_OPEN_MMAPFILE    1
#define CLIENT_OPEN_MMAPFILE    0
typedef struct PATH_MSG_HDR {
    char                version;
    char                operation;
    char                flags;
    volatile char       shm_flags;
    unsigned int        payloadSize;    // in bytes
    uint64_t            startInsnEpoch; // starting instruction epoch of this data blob
    uint64_t            startPathEpoch; // starting path epoch of this data blob
    uint64_t            startStrataEpoch;       // starting strata epoch of this data blob
    char                dataFileBaseName[PATH_MAX];
    } PATH_MSG_HDR;

typedef struct SHARED_MEM {
    struct PATH_MSG_HDR dataHdr;
    char                rawDBBuf[DB_BUF_SIZE];
    } SHARED_MEM;

/*
 * Limitations and assumptions
 * Only one client feeding the shared memory data at a time
 * One approach was tried with sockets, but that was both too slow, and
 * limited by the maximum size of the transmitted content (there was a 64K limit)
 * It is unknown if that 64K limit has been increased subsequent to 2007 when this
 * was being developed.
 *
 * operations:
 *      open_append     append to existing file (default open)
 *      open_trunc      zap any existing file - this is problematic for
 *                      apps which exec without forking (perl test runs)
 *      flush_close     do final writes and close the file
 *      delete          unlink the file after closure
 * Behavior:
 *      sender must issue open type with first blob of data
 *      sender must issue close when sending last blob of data (it can have payload size zero)
 *      the delete flag can be attached to the flush_close op, and if seen, the output files
 *      will be unlinked after closure.
 */
#define OP_OPEN_APPEND          0x1
#define OP_OPEN_TRUNC           0x2
#define OP_OPEN_DEFAULT OP_OPEN_APPEND
#define OP_FLUSH_CLOSE          0x4
#define OP_DELETE               0x8

// Flags: low byte for actions, high byte for shm control
// compression: use libz
// level 6 is a good speed/compression tradeoff; level 9 is max compression
#define FLAG_COMPRESS_6         0x0
#define FLAG_COMPRESS_9         0x1
#define FLAG_NOCOMPRESS         0x2
#define FLAG_COMPRESS_DEFAULT FLAG_COMPRESS_6
#define FLAG_COMPRESS_MAX FLAG_COMPRESS_9

// Clear FILLED and set READY_TO_FILL to tell client it can write data
// Clear READY_TO_FILL and set SHM_FILLED when buffer full, indicating server can read data
#define SHMFLAG_READY_TO_FILL  0x01
#define SHMFLAG_FILLED         0x02

/*
 * Organization of the resultant data:
 *      A text file containing entries:                 <dataFileName>.index
 *              insn epoch:path epoch:offset in the file
 *      A binary file containing the compressed data    <dataFileName>.data
 */
/*
 * server options:  (None of which are yet supported)
 *      -nowrite        do not write to the output file I/O (do compression)
 * Uses a file-backed mmap() approach.
 */
#define STRATA_MMAP_FILE        "/var/tmp/wppp_strata.mmap"
#define COMPRESSOR_MMAP_FILE    "/var/tmp/wppp_compress.mmap"
#define PERM_RW_ALL     0666

#define STATE_NONE      0
#define STATE_OPENED    1
#define STATE_CLOSED    2
typedef struct DF_INFO {
    char        baseFileName[PATH_MAX];
    char        indexFileName[PATH_MAX];
    char        dataFileName[PATH_MAX];
    int         index_fd;
    int         data_fd;
    int         state;
    unsigned short      flags;
    unsigned short      compression;
    off_t       dataFileOffset;
    } DF_INFO;

/*
 * Functions to control the shared memory region
 */

void WaitUntilDataFilled (struct SHARED_MEM* shM);
void MarkDataFilled (struct SHARED_MEM* shM);
void WaitUntilReadyToFill (struct SHARED_MEM* shM);
void MarkReadyToFillData (struct SHARED_MEM* shM);
struct SHARED_MEM* OpenMMapFile(const char *appname, int is_server, const char *mmaped_File);

#endif  // _SERVERCOMM_H_
