/*
 *      Copyright, 2007-2011 (c)  Richard Gorton, All Rights Reserved.
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 *
 *      Author: Rick Gorton     richard.gorton@amd.com
 *
 *      SysV share memory interface for a DWPPP tool
 */
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <zlib.h>

#include <sys/user.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include "servercomm.h"

/*
 * Functions to control the shared memory region
 */

void WaitUntilDataFilled (struct SHARED_MEM* shM)
{
    while (shM->dataHdr.shm_flags != SHMFLAG_FILLED) {
        usleep (1);
    }
// fprintf (stderr, "SHMFLAG_FILLED now set\n");
}

void MarkDataFilled (struct SHARED_MEM* shM)
{
    shM->dataHdr.shm_flags = SHMFLAG_FILLED;
// fprintf (stderr, "SHMFLAG_FILLED now set\n");
}

void WaitUntilReadyToFill (struct SHARED_MEM* shM)
{
    while (shM->dataHdr.shm_flags != SHMFLAG_READY_TO_FILL) {
        usleep (1);
    }
// fprintf (stderr, "SHMFLAG_READY_TO_FILL now set\n");
}

void MarkReadyToFillData (struct SHARED_MEM* shM)
{
    shM->dataHdr.shm_flags = SHMFLAG_READY_TO_FILL;
// fprintf (stderr, "SHMFLAG_READY_TO_FILL now set\n");
}



/*
 * Create an mmap-ed physical file.
 * Doing shmget of a constant key results in a variable shmid, which would require
 * doing other inter-process communication to get the correct shmid... Yuk
 * Returns a pointer to the shared memory region on success; NULL otherwise.
 * Basically, the server creates an empty file to map the SHARED_MEM structure onto.
 * The client simply mmaps to it.
 */
struct SHARED_MEM* OpenMMapFile(const char* opener_app, int is_server, const char *mmap_File)
{
    int shmfd;
    int memSize;
    int rc;
    int flags;
    struct SHARED_MEM*  sharedMem = NULL;

    if (is_server != 0) {
        flags = O_RDWR|O_CREAT|O_EXCL|O_TRUNC;
    } else {
        flags = O_RDWR;
    }
        // Compute the size of the shared region (round up to next page)
    memSize = sizeof (struct SHARED_MEM);
    memSize += PAGE_SIZE;
    memSize = memSize & ~(PAGE_SIZE - 1);

        // Create the mmap file
    shmfd = open (mmap_File, flags, PERM_RW_ALL);
    if (shmfd == -1) {
        fprintf (stderr, "%s: unable to create physical file [%s] to mmap to (errno: %d) - aborting\n",
                opener_app, mmap_File, errno);
        sharedMem = NULL;
        goto done;
    }

    if (is_server != 0) {
        rc = lseek (shmfd, memSize - 1, SEEK_SET);
        if (rc == -1) {
            fprintf (stderr, "%s: unable to seek on mmap-ed file: [%s] (errno: %d) - aborting\n",
                opener_app, mmap_File, errno);
            sharedMem = NULL;
            goto done;
        }
        write (shmfd, "", 1);
        lseek (shmfd, 0, SEEK_SET);
    }

        // And attach to it
    sharedMem = mmap (NULL, memSize, PROT_READ|PROT_WRITE, MAP_SHARED, shmfd, 0);
    close (shmfd);
    if (sharedMem == MAP_FAILED) {
        fprintf (stderr, "%s: unable to mmap the file [%s] (errno: %d) - aborting\n",
                opener_app, mmap_File, errno);
        sharedMem = NULL;
    }

done: ;

    return sharedMem;
}

